# Lokalizace GUI DASv3

Na lokalizaci používáme knihovnu  [GNU gettext](www.gnu.org/software/gettext/). V podstatě slouží pouze k ukládání/načítání katalogu překládaných hlášek, a k jejich překladu za běhu. Kromě toho umí řešit výběr správného plurálu podle jazyka (*Stahuji 1 soubor* vs. *Stahuji 2 soubory* vs. *Stahuji 5 souborů*), i když to jsme (zatím) nepoužili.

## Jak změnit jazyk aplikace

### Pro netrpělivé

Jazyk samotného rozhraní se za normálních okolností bere z operačního systému, tj. pokud máme české windows, používá se čeština, pokud anglické, angličtina atp.. Pokud chcete jazyk nuceně změnit na jiný, odkomentujte v konfiguračním souboru položku ```language.tag``` a nastavte ji na příslušnou hodnotu (*en* pro angličtinu, *cs* pro češtinu, *zh* pro čínštinu atp., seznam všech možností [zde](http://www.iana.org/assignments/language-subtag-registry/language-subtag-registry), hledejte vždy položku **subtag**)

### Podrobnosti

V současnosti se mění jazyk zvlášt pro **uživatelské katalogy** (mění se v konfiguráku, jde o překlady traťových událostí, jednotek a grafů) a zvlášť pro **aplikační katalogy** (bere se z prostředí ve kterém běží, jde o jazyk uživatelského rozhraní a jaderných hlášek). Podrobnosti o uživatelských a aplikačních katalozích jsou v kapitole *Pro překladatele*.

Jazyk aplikačních hlášek se bere z operačního systému, nebo z položky ```language.tag``` v konfiguračním souboru - pokud je uvedena (není zakomentována)

Jazyk uživatelských hlášek se specifikuje nastavením cest k příslušným překladovým souborům v hlavním konfiguráku, tedy položek

 - ```display.units.translation.file```
 - ```plots.translation.file```
 - ```track.events.translation.file```
 
 Cestu k nim je možné uvést "natvrdo", nebo je také možné použít formátovací parametr {language}, který se expanduje na aktuální *language tag* - ať už defaultní z operačního systému, nebo uvedený v konfiguráku. Pomocí parametru {language} je tedy možné snadno zajistit, aby jazyk grafů, jednotek a událostí byl vždy stejný jako jazyk zbytku aplikace, ale zároveň nám ponechává možnost uvést cestu natvrdo, pokud by se to někdy hodilo.
 
 Příklad: 
 
 ```
 plot.translation.file=locale/{language}/LC_MESSAGES/rwtPlots.po
 ```
 
 Poznámka: Parametr {language} je možné uvést i v definici virtuální klávesnice, viz níže.
 

## Jak lokalizace funguje

Lokalizace pracuje se dvěma typy souborů: 

 - Soubory s příponou ```.pot``` (Portable Object Template) - šablony lokalizačních souborů. Obsahují anglickou verzi řetězců k překládání. Vytváří je vývojář - automaticky si je nechává generovat ze zdrojových kódů, a předává je překladateli.
 - Soubory s příponou ```.po``` (Portable Object) jsou již samotné překladové soubory, které vytváří ze souborů ```.pot``` překladatel. Z každé (anglické) šablony ```.pot``` vytvoří nový soubor ```.po``` pro každý jazyk, do kterého se překládá. Tyto soubory se dodávají spolu s hotovu aplikací, je tedy možné do již hotové aplikace přidávat podporu dalších jazyků jen prostým přikopírováním dalších souborů .po
 
Překladatelský proces pak probíhá takto:

 1. Vývojář aplikace vygeneruje šablony souborů ```.pot``` a předá je překladateli.
 2. Překladatel s pomocí vhodné aplikace (např. [Virtaal](http://virtaal.translatehouse.org/) ) vytvoří překlad šablon do příslušného jazyka - výsledkem jsou soubory s překladem .po. Ty se zařadí zpět na příslušné místo do aplikace, která je tím lokalizovaná.
 3. Vývojář vydá novou verzi aplikace - přidal a ubral nějaké řetězce. To znamená, že musí vygenerovat novou šablonu ```.pot``` a předat ji překladateli (1.). Ten ve Virtaalu zvolí merge nové šablony s existujícím překladovým souborem,  dopřeloží nové hlášky, a zařadí aktualizované překlady zpět do aplikace (2.).

## Pro překladatele

Pro práci s lokalizačními soubory existuje několik aplikací pro překladatele, já jednoznačně doporučuji [Virtaal](http://virtaal.translatehouse.org/). Nový překladatel ať si určitě nejprve projede tutorial ve Virtaalu integrovaný, než ho začne používat!! Je mimořádně přínosný, neučí jen práci s aplikací ale i zásadám lokalizace.

Důležité:

  - Je třeba zvolit správný jazyk (ve Virtaalu vpravo dole)
  - Některé hlášky jsou parametrizovatelné, parametry specifikujeme ve složených závorkách s pořadovým číslem. příklad: "River {0} is the {1} longest river in the world." lze přeložit jako "Řeka {0} je {1} nejdelší řeka na světě", ale i "{1} nejdelší řeka na světě je {0}" (všimněte si, že {0} vždy zastupuje název řeky nezávisle na slovosledu!). Aplikace pak do parametrů doplní skutečné hodnoty, takže uživatel uvidí např. "2. nejdelší řeka na světě je Nil". 
  - Co se týče katalogů GUI, překládám je rovnou do češtiny, překladatel má tedy k dispozici i českou i anglickou verzi každé hlášky. Pokud si není jistý významem některé hlášky, nechť ji konzultuje s oběma verzemi překladu.
  - Pokud už byl ```.po``` soubor jednou přeložen, ale poté byla aktualizovaná šablona ```.pot``` (vyšla nová verze aplikace s novými hláškami), Virtaal umí nové hlášky sloučit do již přeloženého souboru. Stačí otevřít již existující ```.po``` a zvolit v menu *File -> Update to template* a vybrat aktualizovanou šablonu.
  
## Co všechno překládat v DASv3 GUI

Katalogy hlášek se dají rozdělit na 

 - **Aplikační katalogy** - generují se ze zdrojových kódů aplikace
 - **Uživatelské katalogy** - např. překlady popisků grafů, traťových událostí, tedy věcí, které při překladu nejsou známy, protože si je uživatel může konfigurovat sám.
 
 Nyní uvedu souhrn toho, co je potřeba přeložit pro aplikaci RWT. Některé věci jsou společné všem aplikacím, jiné (s prefixem rwt) se týkají pouze RWT. U některých aplikací se nepřekládá všechno, například u kalibrační aplikace neexistuje soubor pro překlad popisků grafů, protože tam žádné grafy nejsou.
 
### Aplikační katalogy
 
 šablony jsou umístěné v adresáři ```locale``` projektu. Překlady pro jednotlivé jazyky se pak umísťují do ```locale/en/LC_MESSAGES```, ```locale/cs/LC_MESSAGES```(čeština), ```locale/zh/LC_MESSAGES```(čínština) atp. Seznam zkratek pro všechny jazyky je [zde](http://www.iana.org/assignments/language-subtag-registry/language-subtag-registry), hledejte vždy položku **subtag**
 
 Z Aplikačních katalogů je třeba přeložit
 
  - ```locale/MeasurementCore.pot``` - katalog hlášek měřicího jádra, dodává p. Starzyk. Ostatní šablony aplikačních katalogů dodávám já (Tomáš Dvořák).
  - ```locale/TdScalaCommon.pot``` - překlady mojí knihovny univerzálních komponent, např. popisky virtuální klávesnice, často používané dialogy atp.
  - ```locale/FastScalaPlotsOld.pot``` - překlady mojí knihovny pro vykreslování grafů - např. popisky v popup menu (Fit data, Antialiased rendering atp.)
  - ```locale/DAS.pot``` - hlášky společné všem aplikacím DASGUI
  - ```locale/rwt.pot``` - hlášky specifické pro aplikaci RWT
  
### Uživatelské katalogy
 
 Umístění uživatelských katalogů pro běžící aplikaci se nyní specifikuje v konfiguračním souboru ```config/rwt.cfg```. To znamená, že pro přepnutí překladů např. popisků grafů je třeba změnit cestu k příslušnému překladovému souboru v konfiguračním souboru a restartovat aplikaci. Další možností je použít expanzi pomocí podřetězce {language}, viz kapitola **Jak změnit jazyk aplikace**. Konvence je každopádně taková, že i uživatelské katalogy se umísťují do ```locale/kód_jazyka/LC_MESSAGES``` vedle aplikačních.
 
 Situace u těchto překladových souborů je každopádně o něco složitější. Protože překládají věci, které si může vymýšlet sám uživatel, musí překladatel podle překládaných věcí vytvořit či upravit i šablonu pro překlad (pokud ji už neudělal někdo jiný). Tomu se bohužel z principu nedá vyhnout. Formát šablony je jednoduchý, lze ji upravovat v textovém editoru. Za hlavičkou s metainformacemi (prvních pár řádek) následují dvojice řádků začínající vždy ```msgid``` (hláška k překladu) a ```msgstr``` (přeložená hláška). Každý uživatelský katalog má svá specifika, začneme tím nejjednodušším,
 
#### Uživatelský katalog jednotek
 
 Šablona je v ```locale/DisplayUnits.pot```. Překladový soubor se specifikuje v konfiguračním souboru položkou ```display.units.translation.file```
 
 Používá se v aplikaci i v popiscích os grafu (viz dále), tj. pokud např. chcete u nějaké osy grafu měřítko ve stopách, uvedete tam **feet**, přidáte položku **feet** do šablony, a pomocí Virtaalu otevřete ```locale/cs/LC_MESSAGES/DisplayUnits.po```, dáte **Update to template** na upravenou šablonu, přeložíte **feet** jako **stopy** a uložíte nový katalog. Pokud máte v konfiguračním souboru nastaveno ```display.units.translation.file=locale/cs/LC_MESSAGES/DisplayUnits.po```, uvidíte v popiscích grafu **feet** přeloženo jako **stopy**. Každopádně nepředpokládám, že by se tento soubor upravoval často.
 
#### Uživatelský katalog popisků grafů
 
 Uživatel může vyjít ze šablony ```locale/mmdPlots.pot```. Překladový soubor se specifikuje v konfiguračním souboru položkou ```plots.translation.file```
 
 Ukážu na příkladě grafů aplikace **mmd**, která již má grafy přeložené do češtiny, takže se od ní lze odpíchnout.
 
 Tento katalog překládá popisky grafů, to znamená: Názvy os (```xName```/```yName``` v konfiguráku grafů ```config/mmd.figures.json```), názvy jednotlivých signálů (```signalID``` v konfiguráku grafů).
 
 Protože ruční vytváření šablony tohoto katalogu by bylo mimořádně obtížné, udělal jsem malou utilitku, která ji vytvoří sama z konfiguračního souboru popisovače grafů. Je součástí aplikace DASUtils balené s DASGUI. Stačí spustit DASUtils.exe, stisknout **Convert plot descriptors file into a POT template**, vybrat konfigurační soubor grafů (např. ```config/mmd.plot.descriptors.json```, a aplikace nám ukáže ```.pot``` šablonu. Tu si pak jen zkopírujeme do schránky, uložíme do souboru (např. ```locale/mmdPlots.pot```) a přeložíme jako kteroukoliv jinou šablonu.
 

#### Uživatelský katalog popisků traťových událostí
 
 Šablona tohoto katalogu se generuje pomocí **DASUtils.exe**. Překladový soubor se specifikuje v konfiguračním souboru položkou ```track.events.translation.file```
 
 Uživateli doporučuji podívat se, jak je tento katalog přeložen v aplikaci krtek.
 
 Tento katalog slouží k překladu traťových událostí. Ty samotné se konfigurují v souboru specifikovaném položkou ```track.event.descriptors.file``` v konfiguračním souboru, pro krtka jde např. o soubor ```config/krtek.track.event.descriptors.xml```. Je třeba překládat název traťové události a její parametry, u výčtových hodnot i jednotlivé možnosti výčtu (např. strana Levá/Pravá). 
 
 Protože ruční vytváření šablony tohoto katalogu by bylo mimořádně obtížné, udělal jsem malou utilitku, která ji vytvoří sama z konfiguračního souboru traťových událostí. Je součástí aplikace DASUtils balené s DASGUI. Stačí spustit DASUtils.exe, stisknout **Convert track event descriptors file into a POT template**, vybrat konfigurační soubor traťových událostí (např. ```config/krtek.track.event.descriptors.xml```, a aplikace nám ukáže ```.pot``` šablonu. Tu si pak jen zkopírujeme do schránky, uložíme do souboru (např. ```locale/KrtekTrackEvents.pot```) a přeložíme jako kteroukoliv jinou šablonu.
 
 U traťových událostí mohou být trochu matoucí řetězce k přeložení. U traťové události korekce kilometráže máme např. k přeložení řetězec "CORKM" a řetězec "CorKm". U krtčích událostí se dokonce překládají řetězce jako "1" a "-1". Vysvětlení je následující:
 
 U traťových událostí se v zásadě překládají tři typy hlášek: Symbol (identifikátor) traťové události, identifikátor parametru traťové události, a jednotlivé hodnoty u výčtového parametru události. V komentáři u překládaného řetězce je každopádně v šabloně přibalen vysvětlující komentář, automaticky generovaný z DASUtils Příklad:
 
  - Symbol je např. CORKM nebo PLATFORM, je to vlastně název události, ukazuje se v nabídce událostí nebo jako titulek při vyplňování parametrů události. Komentář v šabloně je např. *Event symbol 'PLATFORM'*
  - Parametr je konkrétní parametr události, například událost "Stožár" (MAST) u krtka má dva parametry, "Length" a "Side". Length je délka v metrech, Side je výčtový typ nabízející dvě možnosti, "-1" a "1". Komentář v šabloně je např. *Parameter 'Length' inside an event 'PLATFORM'*
  - Možnosti výčtového typu jsou např. "-1" a "1 u MAST.Side z předchozího příkladu. Je třeba znát konkrétní význam těchto hodnot, v tomto případě označují levou a pravou stranu, překladatel do angličtiny by je tedy přeložil jako "Left" a "Right". Komentář v šabloně je např. *Value of choice '-1' of parameter 'Side' inside an event 'PLATFORM'*.
 
### Lokalizace virtuální klávesnice
 
 Nejedná se přímo o překlady, ale tematicky se to hodí do tohoto dokumentu. Je možné lokalizovat i rozložení kláves na virtuální klávesnici. Rozložení má svůj vlastní konfigurační soubor, cesta k němu se specifikuje položkou ```virtual.keyboard.file``` v hlavním konfiguračním souboru. Je možné měnit pouze klávesy se znaky, funkční klávesy (šipky, shift atp.) se měnit nedají. Díky tomu je ovšem formát konfiguračního souboru klávesnice triviální, jde jen o dvojice znaků oddělených mezerami - každá dvojice značí znak bez shiftu a se shiftem. Příklad definice standardní US klávesnice:
 
 
    `~ 1! 2@ 3# 4$ 5% 6^ 7& 8* 9( 0) -_ =+
    qQ wW eE rR tT yY uU iI oO pP [{ ]}
    aA sS dD fF gG hH jJ kK lL ;: '" \|
    zZ xX cC vV bB nN mM ,< .> /?


## Pro vývojáře

Šablony (.pot) lze automaticky generovat ze zdrojáků, takže není třeba nic udržovat ručně. Kraťoučké seznámení je [zde](http://cs.wikipedia.org/wiki/GNU_gettext), kompletní manuál [zde](http://www.gnu.org/software/gettext/manual/html_node/gettext.html). Gettext je multiplatformní a podporuje mnoho programovacích jazyků. 

V našem případě hlášky lokalizuje vždy GUI, ať už jsou to přímo stringy UIčka nebo stringy z jádra. Když chce jádro něco lokalizovat, pošle šablonu s placeholdery a zapouzdřené příslušné parametry v protobuf struktuře LocalizableMessage. Překlad provádí UI.

### Jádro

Přehled:

 - Ve zdrojácích se označí části k překladu pomocí ```_("řetezec")```. Označený řetězec slouží zároveň jako klíč do katalogu, a zároveň jako defaultní (fallback) hodnota
 - Pomocí utility ```xgettext``` se ze zdrojáků vygeneruje šablona katalogu (```*.pot```). 
 - Šablona se dá překladatelům, a ti podle ní s pomocí vhodného GUI (virtaal), případně rovnou v textu, udělají soubory (```*.po```) s překlady do jednotlivých jazyků. 

#### Označení řetězů pro automatické generování .pot

My máme výhodu, že jádro nebude dělat žádné překlady samo (pokud by později chtělo, nebude to problém doplnit). Nepotřebujeme tedy v jádře přímo knihovnu gettext používat, jen potřebujeme označit hlášky k překladu gettext-kompatibilním způsobem, aby bylo možné automaticky generovat katalogy hlášek ze zdrojáků. Gettext nepoužívá klíče a hodnoty, hodnota je zároveň klíčem. Technika označení řetězců k překladu je jednoduchá:

```C
#define _(STRING) STRING

printf(_("Hello, world!"))
```

definujeme makro ```_```, které nic nedělá. Nástroj pro tvorbu katalogu ovšem řetězce označené podtržítkem interpretuje jako hlášky k překladu, a vytvoří z nich katalog. Pokud bychom později skutečně chtěli překládat v jádře, změníme jen definici makra na 

```C
#define _(STRING) gettext(string)
```

Generování katalogu z rekurzivního stromu zdrojáků se dá udělat dávkovým souborem s tímto obsahem (dalo mi to chvíli laborování, tak to rovnou vkládám sem):

```
@echo off

echo Generating file list..
dir *.*pp *.h /L /B /S > %TEMP%\listfile.txt

echo Generating .POT file...
"c:\Program Files\GnuWin32\bin\xgettext.exe" -o --force-po --no-wrap --from-code utf-8 -k_ -o locales\default.pot -f %TEMP%\listfile.txt
echo Done.

del %TEMP%\listfile.txt
```
Nejprve je ovšem třeba nainstalovat gettext nástroje, například setupem z [http://gnuwin32.sourceforge.net/packages/gettext.htm]()


#### Komunikace mezi jádrem a GUI

GUI umí překládat řetězce, které mu posílá jádro. Musíme ovšem ještě vymyslet expanzi těchto řetězců podle parametrů, které musíme také dostat. Jádro nám může například chtít poslat hlášku

```
10 files were saved on 12th March, 2020.
```

Ve skutečnosti nám ovšem musí poslat klíč, který se přeloží na hodnotu např.
```
Dne ??? bylo uloženo ??? souborů.
```

a k tomu datum a počet souborů, které si UI teprve expanduje na

```
Dne 12. března bylo uloženo 10 souborů.
```

Máme tedy další dva problémy k řešení:

 - Jaký formát pro expanzi parametrů zvolit a jaké typy parametrů podporovat
 - Jak přenášet pomocí *protobuf* kolekci parametrů.

#### Expanze parametrů

**Poznámka: V praxi nakonec používáme MessageFormat, ale protobuf stále podporuje obě varianty, takže je pro úplnost dokumentuji obě.**

V zásadě se nabízejí dvě možnosti: [printf formát](http://www.pixelbeat.org/programming/gcc/format_specs.html) a [Javovský MessageFormat](http://docs.oracle.com/javase/1.5.0/docs/api/java/text/MessageFormat.html). 

 * printf asi všichni známe. Umí expandovat řetězce, celá a necelá čísla. Umí označovat pozici argumentu. Spec. znaky escapují pomocí zpětného lomítka

    **Příklad:** ```format("Hello, planet n. %1$d on %2$s!", 3, Date.now.toString)``` => *"Hello, planet n. 3 on Sun Apr 06 19:51:37 CEST 2014!"*

    **Varování: Pro celé číslo nepoužívejte ```%i```(integer), ale ```%d```(decimal)**, Javovská implementace ```%i``` neumí.

 * Javovský MessageFormat je určen speciálně pro lokalizace. Umí expandovat celá a necelá čísla, datum, čas, měnu a procenta. U data a času umí vybírat stručnou a dlouhou variantu. Spec. znaky pouze složené závorky, které je možné escapnout uzavřením do apostrofů. Samotný apostrof se pak napíše zdvojením, např. ```"Blockquotes look like this: '{}', and an apostrophe like this:''"``` se přeloží na ```"Blockquotes look like this: {}, and an apostrophe like this:'"```

    **Příklad:** ```format("Hello, planet n. {0} on {1,date}!", 3, Date.now)``` => *"Hello, planet n. 3 on Apr 6, 2014!"*
	
#### Protobuf protokol pro přenášení zpráv

Překladatelná hláška je zapouzdřena ve struktuře ```LocalizableMessage```

```C
/* A gettext-compatible message intended for translation. 

It contains a msgID (in gettext world it is at the same time a key of the message and a fallback value), optionaly a second msgID and a number of items if this message supports a plural form, and a list of parameters for the expansion of the message. 

The expansion method itself is configurable, we can either use a printf syntax, or a java MessageFormat syntax.  The list of the formats can be extended in the future. 

The parameters for the expansion are a collection of a union MessageParameter type. 
*/
message LocalizableMessage { 
	enum FormatSyntax {
		printf = 0; // compatible with java String.format
		messageFormat = 1; // compatible with java MessageFormat
	}

	required FormatSyntax formatSyntax = 1;

	required string msgID = 2;

	// For the plural form both msgID2 and n - the number of items must be filled
	optional string msgID2 = 3;
	optional uint64 n = 4; // It is correct that this int is unsigned, check gettext docs

	// Exactly one member must be filled
	message MessageParameter {
		optional string stringParameter = 1;
		optional int64 intParameter = 2;
		optional double doubleParameter = 3;
		optional int64 timestampParameter = 4; // Unix epoch time*1000. time(NULL)*1000 in C or System.currentTimeMillis() in Java.
	}

	// parameteres for message expansion
	repeated MessageParameter parameters = 5;
}
```

### GUI

Nevím jak je to s ostatními programovacími jazyky, já píšu GUI ve scale. .pot generuji před každým release pomocí skriptu ```run_xgettext.sh``` v kořenovém adresáři projektu.
